###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import operator
from collections import defaultdict
import multiprocessing as mp

import biolib.seq_io as seq_io
from biolib.taxonomy import Taxonomy

from genometreetk.exceptions import GenomeTreeTkError
from genometreetk.common import (read_gtdb_taxonomy,
                                 read_gtdb_ncbi_taxonomy,
                                 read_gtdb_ncbi_organism_name,
                                 species_label,
                                 predict_bacteria,
                                 check_domain_assignment,
                                 assign_rep)


class Cluster(object):
    """Cluster genomes based on AAI of concatenated alignment.

    Genomes are preferentially assigned to representatives from
    RefSeq, followed by GenBank, and final User genomes. This
    is done to ensure the majority of genomes are assigned to
    publicly available representatives. However, this means
    a genome will not necessarily be assigned to the representative
    with the highest AAI. Furthermore, genomes are not clustered
    if they have different valid species names. NCBI genomes are
    never assigned to User representatives to ensure they always
    appear in trees without User genomes.
    """

    def __init__(self, cpus):
        """Initialization.

        Parameters
        ----------
        cpus : int
            Maximum number of cpus/threads to use.
        """

        self.logger = logging.getLogger()

        self.cpus = cpus

    def __worker(self,
                 representatives,
                 bac_seqs,
                 ar_seqs,
                 aai_threshold,
                 metadata_file,
                 trusted_user_genomes,
                 queue_in,
                 queue_out):
        """Process genomes in parallel."""
        
        gtdb_taxonomy = read_gtdb_taxonomy(metadata_file)
        ncbi_taxonomy = read_gtdb_ncbi_taxonomy(metadata_file)

        # read taxonomy information and determine 'best' species label for each genome
        ncbi_organism_names = read_gtdb_ncbi_organism_name(metadata_file)
        species = species_label(gtdb_taxonomy, ncbi_taxonomy, ncbi_organism_names)

        # determine genus of all genomes and representatives
        genus = {}
        reps_from_genus = defaultdict(set)
        for genome_id, t in gtdb_taxonomy.iteritems():
            g = None
            if len(t) >= 6 and t[5] != 'g__':
                g = t[5]
            elif genome_id in species:
                g = species[genome_id].split(' ')[0][3:]

            if g:
                genus[genome_id] = g

                if genome_id in representatives:
                    reps_from_genus[g].add(genome_id)
                    
        # predict domain of each representative
        rep_is_bacteria = {}
        for rep_id in representatives:
            rep_is_bacteria[rep_id], per_bac_aa, per_ar_aa = predict_bacteria(rep_id, bac_seqs, ar_seqs)
            if not check_domain_assignment(rep_id, gtdb_taxonomy, ncbi_taxonomy, rep_is_bacteria[rep_id]):
                print 'Problem with representative.'
                print 'Bac vs Ar:', per_bac_aa, per_ar_aa

        # assign representatives           
        while True:
            genome_id = queue_in.get(block=True, timeout=None)
            if genome_id == None:
                break
                
            cur_aai = aai_threshold
            assigned_representative = None

            genome_is_bacteria, per_bac_aa, per_ar_aa = predict_bacteria(genome_id, bac_seqs, ar_seqs)
            if not check_domain_assignment(genome_id, gtdb_taxonomy, ncbi_taxonomy, genome_is_bacteria):
                print 'Bac vs Ar:', per_bac_aa, per_ar_aa
            else:
                if genome_is_bacteria:
                    genome_seq = bac_seqs[genome_id]
                else:
                    genome_seq = ar_seqs[genome_id]

                genome_aa_count = len(genome_seq) - genome_seq.count('-')
                min_matches = max(0.5*genome_aa_count, 0.1*len(genome_seq))

                # speed up computation by first comparing genome 
                # to representatives of the same genus
                genome_genus = genus.get(genome_id, None)
                cur_reps_from_genus = reps_from_genus.get(genome_genus, set())
                remaining_reps = representatives.difference(cur_reps_from_genus)
                for rep_set in [cur_reps_from_genus, remaining_reps]:
                    for rep_id in rep_set:
                        assigned_representative, cur_aai = assign_rep(rep_id, 
                                                                        genome_id,
                                                                        rep_is_bacteria[rep_id], 
                                                                        genome_is_bacteria,
                                                                        bac_seqs, 
                                                                        ar_seqs,
                                                                        species,
                                                                        gtdb_taxonomy,
                                                                        genome_aa_count,
                                                                        trusted_user_genomes,
                                                                        aai_threshold,
                                                                        min_matches,
                                                                        assigned_representative,
                                                                        cur_aai)

            queue_out.put((genome_id, assigned_representative, cur_aai))

    def __writer(self, representatives, metadata_file, num_genomes, output_file, writer_queue):
        """Process representative assignments from each process.

        Parameters
        ----------
        representatives : set
            Initial set of representative genomes.
        metadata_file : str
            GTDB metadata file.
        num_genomes : int
            Number of genomes being processed.
        output_file : str
            Output file specifying genome clustering.
        """
        
        ncbi_taxonomy = read_gtdb_ncbi_taxonomy(metadata_file)

        # initialize clusters
        clusters = {}
        for rep_id in representatives:
            clusters[rep_id] = []

        # gather results for each genome
        fout_details = open(output_file + '.details', 'w')
        fout_details.write('Representative Id\tGenome Id\tAAI\tRep. NCBI taxonomy\tGenome NCBI Taxonomy\n')
        processed_genomes = 0
        while True:
            genome_id, assigned_representative, aai = writer_queue.get(block=True, timeout=None)
            if genome_id == None:
              break

            processed_genomes += 1
            statusStr = 'Finished processing %d of %d (%.2f%%) genomes.' % (processed_genomes,
                                                                            num_genomes,
                                                                            float(processed_genomes) * 100 / num_genomes)
            sys.stdout.write('%s\r' % statusStr)
            sys.stdout.flush()

            if assigned_representative:
                clusters[assigned_representative].append(genome_id)
                
                fout_details.write('%s\t%s\t%.2f\t%s\t%s\n' % (assigned_representative,
                                                                genome_id,
                                                                aai * 100,
                                                                ';'.join(ncbi_taxonomy.get(assigned_representative, Taxonomy.rank_prefixes)),
                                                                ';'.join(ncbi_taxonomy.get(genome_id, Taxonomy.rank_prefixes))))

        sys.stdout.write('\n')
        fout_details.close()

        # write out clusters
        fout = open(output_file, 'w')
        clustered_genomes = 0
        for c, cluster_rep in enumerate(sorted(clusters, key=lambda x: len(clusters[x]), reverse=True)):
            cluster_str = 'cluster_%d' % (c + 1)
            cluster = clusters[cluster_rep]
            clustered_genomes += len(cluster)
            fout.write('%s\t%s\t%d\t%s\n' % (cluster_rep, cluster_str, len(cluster) + 1, ','.join(cluster)))

        fout.close()

        self.logger.info('Assigned %d genomes to representatives.' % clustered_genomes)

    def _cluster(self,
                    representatives,
                    genomes_to_process,
                    aai_threshold,
                    ar_seqs,
                    bac_seqs,
                    metadata_file,
                    trusted_user_genomes,
                    output_file):
        """Assign genomes to representatives.


        Parameters
        ----------
        representatives : set
            Initial set of representative genomes.
        genomes_to_process : list
            Genomes to process for identification of new representatives.
        aai_threshold : float
              AAI threshold for assigning a genome to a representative.
        ar_seqs : d[genome_id] -> alignment
            Alignment of archaeal marker genes.
        bac_seqs : d[genome_id] -> alignment
            Alignment of bacterial marker genes.
        metadata_file : str
            Metadata, including CheckM estimates, for all genomes.
        trusted_user_genomes : set
            Trusted User genomes to treat as if they were in GenBank.
        output_file : str
            Output file specifying genome clustering.
        """

        # populate worker queue with data to process
        worker_queue = mp.Queue()
        writer_queue = mp.Queue()

        for genome_id in genomes_to_process:
          worker_queue.put(genome_id)

        for _ in range(self.cpus):
          worker_queue.put(None)

        try:
          worker_proc = [mp.Process(target=self.__worker, args=(representatives,
                                                                    bac_seqs,
                                                                    ar_seqs,
                                                                    aai_threshold,
                                                                    metadata_file,
                                                                    trusted_user_genomes,
                                                                    worker_queue,
                                                                    writer_queue)) for _ in range(self.cpus)]
          write_proc = mp.Process(target=self.__writer, args=(representatives,
                                                                metadata_file,
                                                                  len(genomes_to_process),
                                                                  output_file,
                                                                  writer_queue))

          write_proc.start()

          for p in worker_proc:
              p.start()

          for p in worker_proc:
              p.join()

          writer_queue.put((None, None, None))
          write_proc.join()
        except:
          for p in worker_proc:
            p.terminate()

          write_proc.terminate()

    def run(self,
            representatives_file,
            trusted_user_file,
            ar_msa_file,
            bac_msa_file,
            aai_threshold,
            metadata_file,
            output_file):
        """Identify additional representatives based on AAI between aligned sequences.

        Parameters
        ----------
        representatives_file : str
            File listing genome identifiers as initial representatives.
        trusted_user_file : str
            File listing trusted User genomes that should be treated as if they are in GenBank.
        ar_msa_file : str
            Name of file containing canonical archaeal multiple sequence alignment.
        bac_msa_file : str
            Name of file containing canonical bacterial multiple sequence alignment.
        aai_threshold : float
              AAI threshold for clustering genomes to a representative.
        metadata_file : str
            Metadata, including CheckM estimates, for all genomes.
        output_file : str
            Output file specifying genome clustering.
        """

        # read sequences
        ar_seqs = seq_io.read_fasta(ar_msa_file)
        bac_seqs = seq_io.read_fasta(bac_msa_file)
        self.logger.info('Identified %d archaeal sequences in MSA.' % len(ar_seqs))
        self.logger.info('Identified %d bacterial sequences in MSA.' % len(bac_seqs))

        if len(ar_seqs) != len(bac_seqs):
            self.logger.error('Archaeal and bacterial MSA files do not contain the same number of sequences.')
            raise GenomeTreeTkError('Error with MSA input files.')

        genome_to_consider = set(ar_seqs.keys())

        # read initial representatives
        rep_genomes = set()
        for line in open(representatives_file):
            if line[0] == '#':
                continue

            genome_id = line.rstrip().split('\t')[0]
            if genome_id not in genome_to_consider:
                self.logger.error('Representative genome %s has no sequence data.' % genome_id)
            rep_genomes.add(genome_id)

        self.logger.info('Identified %d representatives.' % len(rep_genomes))
        
        # read trusted User genomes
        trusted_user_genomes = set()
        if trusted_user_file:
            for line in open(trusted_user_file):
                line_split = line.split('\t')
                trusted_user_genomes.add(line_split[0])
                
        self.logger.info('Identified %d trusted User genomes.' % len(trusted_user_genomes)) 

        # cluster genomes to representatives
        genomes_to_cluster = genome_to_consider - rep_genomes
        self.logger.info('Comparing %d genomes to %d representatives with threshold = %.3f.' % (len(genomes_to_cluster),
                                                                                                len(rep_genomes),
                                                                                                aai_threshold))
        self._cluster(rep_genomes,
                        genomes_to_cluster,
                        aai_threshold,
                        ar_seqs,
                        bac_seqs,
                        metadata_file,
                        trusted_user_genomes,
                        output_file)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
from collections import defaultdict

from math import floor

import dendropy

from biolib.common import is_float
from biolib.taxonomy import Taxonomy


class PhylogeneticDiversity():
    """Calculate phylogenetic diversity."""

    def __init__(self):
        """Initialize."""
        self.logger = logging.getLogger()
        
    def _read_taxa_list(self, taxa_list):
        """Read taxa from file."""
        
        taxa = set()
        for line in open(taxa_list):
            taxa.add(line.strip().split('\t')[0])
            
        return taxa
        
    def _total_pd(self, tree):
        """Calculate PD over entire tree."""
        
        total_pd = 0
        total_taxa = 0
        for node in tree.preorder_node_iter():
            if node.parent_node is not None:
                total_pd += node.edge.length
                if node.is_leaf():
                    total_taxa += 1
                    
        return total_pd, total_taxa
        
    def _read_reps(self, rep_list):
        """Read genomes assigned to a representative."""
        
        genome_reps = {}
        
        if rep_list:
            for line in open(rep_list):
                line_split = line.strip().split('\t')
                
                rep_id = line_split[0]
                
                if len(line_split) >= 2:
                    for genome_id in line_split[1].split(';'):
                        genome_reps[genome_id] = rep_id
                
        return genome_reps
        
    def _include_reps(self, ingroup_taxa, outgroup_taxa, genome_reps, is_ingroup):
        """Expand set of taxa to include representatives of taxa."""
        
        rep_is_ingroup = []
        rep_is_outgroup = []
        if is_ingroup:
            taxa_with_reps = set(ingroup_taxa)
            for genome_id in ingroup_taxa:
                if genome_id in genome_reps:
                    rep_id = genome_reps[genome_id]
                    if rep_id == genome_id:
                        continue
                        
                    taxa_with_reps.add(rep_id)
                    if rep_id in ingroup_taxa:
                        rep_is_ingroup.append(rep_id)
                    else:
                        rep_is_outgroup.append(rep_id)
        else:
            taxa_with_reps = set(outgroup_taxa)
            for genome_id in genome_reps:
                if genome_id in ingroup_taxa:
                    continue
                    
                rep_id = genome_reps[genome_id]
                if rep_id == genome_id:
                    continue
                    
                if rep_id in ingroup_taxa:
                    taxa_with_reps.add(rep_id)
                    rep_is_ingroup.append(rep_id)
                elif rep_id in outgroup_taxa:
                    taxa_with_reps.add(rep_id)
                    rep_is_outgroup.append(rep_id)
                else:
                    # this should never happend
                    self.logger.warning('There is an outgroup taxa (%s) with a representative (%s) not present in the tree.' % (genome_id, rep_id))
                        
        return taxa_with_reps, rep_is_ingroup, rep_is_outgroup
        
    def _taxon_pd(self, tree, ingroup, out_taxa_with_reps, genome_reps):
        """Calculate phylogenetic gain of each ingroup taxon relative to outgroup."""

        pg_taxon = {}
        for taxon in ingroup:
            rep_id = genome_reps.get(taxon, None)
            if rep_id and rep_id not in ingroup:
                pg_taxon[taxon] = [0, rep_id + ' (assigned to outgroup representative)']
            
        for leaf in tree.leaf_node_iter():
            if leaf.taxon.label in ingroup:
                # find first internal node containing an outgroup taxon
                pg = 0
                parent = leaf
                outgroup_taxon = 'None'
                while parent:
                    # check for outgroup taxon
                    stop = False
                    for tip in parent.leaf_iter():
                        if tip.taxon.label in out_taxa_with_reps:
                            outgroup_taxon = tip.taxon.label
                            if outgroup_taxon in ingroup:
                                outgroup_taxon += ' (one or more outgroup taxa are assigned to this ingroup taxon)'
                            stop = True
                            
                    if stop:
                        break
                    
                    pg += parent.edge.length
                    parent = parent.parent_node
                    
                pg_taxon[leaf.taxon.label] = [pg, outgroup_taxon]
                
                # propagate information to genomes represented by this genome_id
                for genome_id, rep_id in genome_reps.iteritems():
                    if genome_id in ingroup and rep_id == leaf.taxon.label:
                        pg_taxon[genome_id] = [pg, outgroup_taxon]

        return pg_taxon

    def pd(self, tree, taxa_list, rep_list, per_taxa_pg_file):
        """Calculate phylogenetic diversity of extant taxa."""

        genome_reps = self._read_reps(rep_list)
        
        self.logger.info('Reading tree.')
        tree = dendropy.Tree.get_from_path(tree,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)
                                                                            
        # get total branch length of tree
        self.logger.info('Calculating total PD.')
        total_pd, total_taxa = self._total_pd(tree)

        # get PD of ingroup taxa
        self.logger.info('Calculating total PD for specified taxa.')
        ingroup = self._read_taxa_list(taxa_list)
        
        in_taxa = set()
        for leaf in tree.leaf_node_iter():
            if leaf.taxon.label in ingroup:
                in_taxa.add(leaf.taxon.label)
            
        #tree_in = dendropy.Tree(tree)
        #tree_in.retain_taxa_with_labels(ingroup)
        #in_pd, in_taxa = self._total_pd(tree_in)
        
        self.logger.info('Specified ingroup taxa: %d' % len(ingroup))
        self.logger.info('Ingroup taxa as representatives or singletons in tree: %d' % len(in_taxa))

        # calculate PD for ingroup with additional genomes assigned to a representative
        ingroup_with_reps, rep_is_ingroup, rep_is_outgroup = self._include_reps(ingroup, None, genome_reps, True)
        tree_in_with_reps = dendropy.Tree(tree)
        tree_in_with_reps.retain_taxa_with_labels(ingroup_with_reps)
        in_pd_with_reps, in_taxa_with_reps = self._total_pd(tree_in_with_reps)
        
        self.logger.info('Ingroup taxa represented by another ingroup taxa: %d' % len(rep_is_ingroup))
        self.logger.info('Ingroup taxa represented by an outgroup taxa: %d' % len(rep_is_outgroup))
        self.logger.info('Unique outgroup taxa representing an ingroup taxa: %d' % len(set(rep_is_outgroup)))
        ingroup_taxa = len(in_taxa) + len(rep_is_ingroup) + len(rep_is_outgroup)
        ingroup_taxa_derep = len(in_taxa) + len(set(rep_is_outgroup))
        
        # get PD of outgroup taxa
        outgroup = set()
        for leaf in tree.leaf_node_iter():
            if leaf.taxon.label not in ingroup:
                outgroup.add(leaf.taxon.label)
                
        self.logger.info('Outgroup taxa as representatives or singletons in tree: %d' % len(outgroup))
        
        # calculate PD for outgroup with additional genomes assigned to a representative
        #tree_out = dendropy.Tree(tree)
        #tree_out.retain_taxa_with_labels(outgroup)
        #out_pd, out_taxa = self._total_pd(tree_out)

        outgroup_with_reps, rep_is_ingroup, rep_is_outgroup = self._include_reps(ingroup, outgroup, genome_reps, False)
        tree_out_with_reps = dendropy.Tree(tree)
        tree_out_with_reps.retain_taxa_with_labels(outgroup_with_reps)
        out_pd_with_reps, out_taxa_with_reps = self._total_pd(tree_out_with_reps)
                
        self.logger.info('Outgroup taxa represented by another outgroup taxa: %d' % len(rep_is_outgroup))
        self.logger.info('Outgroup taxa represented by an ingroup taxa: %d' % len(rep_is_ingroup))
        self.logger.info('Unique ingroup taxa representing an outgroup taxa: %d' % len(set(rep_is_ingroup)))
        outgroup_taxa = len(outgroup) + len(rep_is_outgroup) + len(rep_is_ingroup)
        outgroup_taxa_derep = len(outgroup) + len(set(rep_is_ingroup))
        
        # calculate PG of each ingroup taxon relative to outgroup in requested
        if per_taxa_pg_file:
            self.logger.info('Calculating PG of each ingroup taxon relative to outgroup.')
            pg_taxon = self._taxon_pd(tree, ingroup, outgroup.union(rep_is_ingroup), genome_reps)
            
            fout = open(per_taxa_pg_file, 'w')
            fout.write('Taxon\tPG\tPercent PG\tFirst outgroup taxon\n')
            for taxon, pg_stats in pg_taxon.iteritems():
                pg, outgroup_taxon = pg_stats
                fout.write('%s\t%f\t%f\t%s\n' % (taxon, pg, pg * 100.0 / total_pd, outgroup_taxon))
            fout.close()
                   
        return total_pd, ingroup_taxa, ingroup_taxa_derep, in_pd_with_reps, outgroup_taxa, outgroup_taxa_derep, out_pd_with_reps
        
    def _clade_pd(self, tree, ingroup_count, outgroup_count):
        """Calculate PD for named clades."""
        
        pd = {}
        for node in tree.preorder_node_iter():
            if not node.label:
                continue

            taxon = None
            if ':' in node.label:
                _support, taxon = node.label.split(':')
            else:
                if not is_float(node.label):
                    taxon = node.label

            if taxon:
                taxon_pd = 0
                taxon_count = 0
                in_taxon_pd = 0
                in_taxon_count = 0
                in_taxon_derep = 0
                out_taxon_pd = 0
                out_taxon_count = 0
                out_taxon_derep = 0
                for nn in node.postorder_iter():
                    if nn == node:
                        continue
                        
                    # check if group contains taxa from
                    # the ingroup and/or outgroup
                    ingroup_leaves = False
                    outgroup_leaves = False
                    for leaf in nn.leaf_iter():
                        genome_id = leaf.taxon.label
                        if genome_id in ingroup_count:
                            ingroup_leaves = True
                        
                        if genome_id in outgroup_count:
                            outgroup_leaves = True
                            
                    if ingroup_leaves:
                        in_taxon_pd += nn.edge.length

                    if outgroup_leaves:
                        out_taxon_pd += nn.edge.length
                        
                    if nn.is_leaf():
                        genome_id = nn.taxon.label
                        
                        in_taxon_count += ingroup_count.get(genome_id, 0)
                        if genome_id in ingroup_count:
                            in_taxon_derep += 1
                            
                        out_taxon_count += outgroup_count.get(genome_id, 0)
                        if genome_id in outgroup_count:
                            out_taxon_derep += 1
                            
                    taxon_pd += nn.edge.length

                if taxon == 'd__Archaea':
                    print taxon_count, in_taxon_count, out_taxon_count, in_taxon_pd, out_taxon_pd
                pd[taxon] = [taxon_pd, in_taxon_pd, in_taxon_count, in_taxon_derep, out_taxon_pd, out_taxon_count, out_taxon_derep]
                
        return pd
        
    def pd_clade(self, decorated_tree, output_file, taxa_list, rep_list):
        """Calculate phylogenetic diversity of named groups."""
        
        # calculate PD for entire tree
        self.logger.info('Reading tree.')
        tree = dendropy.Tree.get_from_path(decorated_tree,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)
 
        self.logger.info('Calculating total PD.')
        total_pd, total_taxa = self._total_pd(tree)
        
        # get ingroup and outgroup taxa
        ingroup = self._read_taxa_list(taxa_list)
        self.logger.info('Specified ingroup taxa: %d' % len(ingroup))
        
        # get number of ingroup and outgroup genomes 
        # representated by each leaf node
        genome_reps = self._read_reps(rep_list)
        
        ingroup_count = defaultdict(int)
        outgroup_count = defaultdict(int)
        for genome_id in genome_reps:
            rep_id = genome_reps[genome_id]
                
            if genome_id in ingroup:
                ingroup_count[rep_id] += 1
            else:
                outgroup_count[rep_id] += 1
                
        # add count for singletons
        for leaf in tree.leaf_node_iter():
            genome_id = leaf.taxon.label
            
            if genome_id in ingroup and genome_id not in ingroup_count:
                ingroup_count[genome_id] = 1 # ingroup singleton
            elif genome_id not in ingroup and genome_id not in  outgroup_count:
                outgroup_count[genome_id] = 1 # outgroup singleton

        # PD for named groups
        self.logger.info('Calculating PD for named clades.')
        pd_clade = self._clade_pd(tree, ingroup_count, outgroup_count)
        
        print 'ingroup_count, outgroup_count', len(ingroup_count), len(outgroup_count)

        # report results
        fout = open(output_file, 'w')
        fout.write('Clade')
        fout.write('\tTaxa\tTaxa (derep)\tPD\tPercent PD')
        fout.write('\tOut Taxa\tOut Taxa (derep)\tOut PD\tOut Percent PD')
        fout.write('\tIn Taxa\tIn Taxa (derep)\tIn PD\tIn Percent PD')
        fout.write('\tIn PG\tIn Percent PG\n')
        
        ordered_taxa = Taxonomy().sort_taxa(pd_clade.keys())
        for taxon in ordered_taxa:
            taxon_pd, in_taxon_pd, in_taxon_count, in_taxon_derep, out_taxon_pd, out_taxon_count, out_taxon_derep = pd_clade[taxon]
            taxon_count = in_taxon_count + out_taxon_count
            taxon_derep = in_taxon_derep + out_taxon_derep
            in_taxon_pg = taxon_pd - out_taxon_pd
            
            taxon_pd = max(taxon_pd, 1e-9) # make sure PD is never exactly zero to avoid division errors
            
            row = taxon
            row += '\t%d\t%d\t%.2f\t%.2f' % (taxon_count, taxon_derep, taxon_pd, taxon_pd * 100 / total_pd)
            row += '\t%d\t%d\t%.2f\t%.2f' % (out_taxon_count, out_taxon_derep, out_taxon_pd, out_taxon_pd * 100 / taxon_pd)
            row += '\t%d\t%d\t%.2f\t%.2f' % (in_taxon_count, in_taxon_derep, in_taxon_pd, in_taxon_pd * 100 / taxon_pd)
            row += '\t%.2f\t%.2f' % (in_taxon_pg, in_taxon_pg * 100 / taxon_pd)
            fout.write(row + '\n')
        fout.close()
